import requests
import base64
import json
import logging
import time

from O365.event import Event

log = logging.getLogger(__name__)

class Calendar( object ):
	'''
	Calendar manages lists of events on an associated calendar on office365.
	
	Methods:
		getName - Returns the name of the calendar.
		getCalendarId - returns the GUID that identifies the calendar on office365
		getId - synonym of getCalendarId
		getEvents - kicks off the process of fetching events.
		fetchEvents - legacy duplicate of getEvents
	
	Variable:
		events_url - the url that is actually called to fetch events. takes an ID, start, and end date.
		time_string - used for converting between struct_time and json's time format.
	'''
	events_url = 'https://outlook.office365.com/api/v1.0/me/calendars/{0}/calendarview?startDateTime={1}&endDateTime={2}&$top={3}'
	time_string = '%Y-%m-%dT%H:%M:%SZ'

	def __init__(self, json=None, auth=None, verify=True):
		'''
		Wraps all the information for managing calendars.
		'''
		self.json = json
		self.auth = auth
		self.events = []

		if json:
			log.debug('translating calendar information into local variables.')
			self.calendarId = json['Id']
			self.name = json['Name']

		self.verify = verify

	def getName(self):
		'''Get the calendar's Name.'''
		return self.json['Name']

	def getCalendarId(self):
		'''Get calendar's GUID for office 365. mostly used internally in this library.'''
		return self.json['Id']

	def getId(self):
		'''Get calendar's GUID for office 365. mostly used internally in this library.'''
		return self.getCalendarId()

	def fetchEvents(self,start=None,end=None):
		'''
		So I originally made this function "fetchEvents" which was a terrible idea. Everything else
		is "getX" except events which were appearenty to good for that. So this function is just a 
		pass through for legacy sake.
		'''
		return self.getEvents(start,end)


	def getEvents(self,start=None,end=None, eventCount=10):
		'''
		Pulls events in for this calendar. default range is today to a year now.
		
		Keyword Arguments:
		start -- The starting date from where you want to begin requesting events. The expected 
		type is a struct_time. Default is today.
		end -- The ending date to where you want to end requesting events. The expected 
		type is a struct_time. Default is a year from start.
		'''

		# If no start time has been supplied, it is assumed you want to start as of now.
		if not start:
			start = time.strftime(self.time_string)

		# If no end time has been supplied, it is assumed you want the end time to be a year
		# from what ever the start date was. 
		if not end:
			end = time.time()
			end += 3600*24*365
			end = time.gmtime(end)
			end = time.strftime(self.time_string,end)

		# This is where the actual call to Office365 happens.
		response = requests.get(self.events_url.format(self.json['Id'],start,end,eventCount) ,auth=self.auth, verify=self.verify)
		log.info('Response from O365: %s', str(response))
		
		#This takes that response and then parses it into individual calendar events.
		for event in response.json()['value']:
			try:
				duplicate = False

				# checks to see if the event is a duplicate. if it is local changes are clobbered.
				for i,e in enumerate(self.events):
					if e.json['Id'] == event['Id']:
						self.events[i] = Event(event,self.auth,self)
						duplicate = True
						break

				if not duplicate:
					self.events.append(Event(event,self.auth,self))
				
				log.debug('appended event: %s',event['Subject'])
			except Exception as e:
				log.info('failed to append calendar: %',str(e))
		
		log.debug('all events retrieved and put in to the list.')
		return True

# To the King!



import logging

from O365.connection import Connection
from O365.fluent_message import Message

log = logging.getLogger(__name__)


class FluentInbox(object):
    url_dict = {
        'inbox': {
            '1.0': 'https://outlook.office365.com/api/v1.0/me/messages',
            '2.0': 'https://graph.microsoft.com/v1.0/me/messages',
        },

        'folders': {
            '1.0': 'https://outlook.office365.com/api/v1.0/me/Folders',
            '2.0': 'https://graph.microsoft.com/v1.0/me/MailFolders',
        },

        'folder': {
            '1.0': 'https://outlook.office365.com/api/v1.0/me/Folders/{folder_id}/messages',
            '2.0': 'https://graph.microsoft.com/v1.0/me/MailFolders/{folder_id}/messages',
        },
    }

    def __init__(self, verify=True):
        """ Creates a new inbox wrapper.

        :param verify: whether or not to verify SSL certificate
        """
        self.url = FluentInbox._get_url('inbox')
        self.fetched_count = 0
        self._filter = ''
        self._search = ''
        self.verify = verify
        self.messages = []

    def from_folder(self, folder_name):
        """ Configure to use this folder for fetching the mails

        :param folder_name: name of the outlook folder
        """
        self._reset()
        response = Connection.get_response(FluentInbox._get_url('folders'),
                                           verify=self.verify,
                                           params={'$top': 100})

        folder_id = None
        all_folders = []

        for folder in response:
            if folder['displayName'] == folder_name:
                folder_id = folder['id']
                break

            all_folders.append(folder['displayName'])

        if not folder_id:
            raise RuntimeError('Folder "{}" is not found, available folders '
                               'are {}'.format(folder_name, all_folders))

        self.url = FluentInbox._get_url('folder').format(folder_id=folder_id)

        return self

    def filter(self, filter_string):
        """ Set the value of a filter. More information on what filters are available can be found here:
        https://msdn.microsoft.com/office/office365/APi/complex-types-for-mail-contacts-calendar#RESTAPIResourcesMessage
        More improvements coming soon

        :param filter_string: The string that represents the filters you want to enact.
                should be something like: (HasAttachments eq true) and (IsRead eq false) or just: IsRead eq false
                test your filter string here: https://outlook.office365.com/api/v1.0/me/messages?$filter=
                if that accepts it then you know it works.
        """
        self._filter = filter_string
        return self

    def search(self, search_string):
        """ Set the value of a search. More information on what searches are available can be found here:
        https://msdn.microsoft.com/office/office365/APi/complex-types-for-mail-contacts-calendar#RESTAPIResourcesMessage
        More improvements coming soon

        :param search_string: The search string you want to use

        Should be something like: "Category:Action AND Subject:Test" or just: "Subject:Test".

        Test your search string here: "https://outlook.office365.com/api/v1.0/me/messages?$search="
        or directly in your mailbox, if that accepts it then you know it works.
        """
        self._search = search_string
        return self

    def fetch_first(self, count=10):
        """ Fetch the first n messages, where n is the specified count

        :param count: no.of messages to fetch
        """
        self.fetched_count = 0
        return self.fetch_next(count=count)

    def skip(self, count):
        """ Skips the first n messages, where n is the specified count

        :param count: no.of messages to skip
        """
        self.fetched_count = count
        return self

    def fetch(self, count=10):
        """ Fetch n messages from the result, where n is the specified count

        :param count: no.of messages to fetch
        """
        return self.fetch_next(count=count)

    def fetch_next(self, count=1):
        """ Fetch the next n messages after the previous fetch, where n is the specified count

        :param count: no.of messages to fetch
        """
        skip_count = self.fetched_count
        if self._search:
            params = {'$filter': self._filter, '$top': count,
                      '$search': '"{}"'.format(self._search)}
        else:
            params = {'$filter': self._filter, '$top': count,
                      '$skip': skip_count}

        response = Connection.get_response(self.url, verify=self.verify,
                                           params=params)
        self.fetched_count += count

        messages = []
        for message in response:
            messages.append(Message(message, Connection().auth))

        return messages

    @staticmethod
    def _get_url(key):
        """ Fetches the url for specified key as per the connection version configured

        :param key: the key for which url is required
        :return: URL to use for requests
        """
        return FluentInbox.url_dict[key][Connection().api_version]

    def _reset(self):
        """ Resets the current reference """
        self.fetched_count = 0
        self.messages = []



import logging
import os
import os.path as path
import json

import requests
from oauthlib.oauth2 import TokenExpiredError
from requests_oauthlib import OAuth2Session
from future.utils import with_metaclass


log = logging.getLogger(__name__)


class MicroDict(dict):
    def __getitem__(self, key):
        result = super(MicroDict, self).get(key[:1].lower() + key[1:], None)
        if result is None:
            result = super(MicroDict, self).get(key[:1].upper() + key[1:])
        return result


class Singleton(type):
    _instance = None

    def __call__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instance

    # def __new__(cls, *args, **kwargs):
    #     if not cls._instance:
    #         cls._instance = object.__new__(cls)
    #     return cls._instance


_default_token_file = '.o365_token'
_home_path = path.expanduser("~")
default_token_path = path.join(_home_path, _default_token_file)


def save_token(token, token_path=None):
    """ Save the specified token dictionary to a specified file path

    :param token: token dictionary returned by the oauth token request
    :param token_path: path to where the files is to be saved
    """
    if not token_path:
        token_path = default_token_path

    with open(token_path, 'w') as token_file:
        json.dump(token, token_file, indent=True)


def load_token(token_path=None):
    """ Save the specified token dictionary to a specified file path

    :param token_path: path to the file with token information saved
    """
    if not token_path:
        token_path = default_token_path

    token = None
    if path.exists(token_path):
        with open(token_path, 'r') as token_file:
            token = json.load(token_file)
    return token


def delete_token(token_path=None):
    """ Save the specified token dictionary to a specified file path

    :param token_path: path to where the token is saved
    """
    if not token_path:
        token_path = default_token_path

    if path.exists(token_path):
        os.unlink(token_path)


class Connection(with_metaclass(Singleton)):
    _oauth2_authorize_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/authorize'
    _oauth2_token_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/token'

    def __init__(self):
        """ Creates a O365 connection object """
        self.api_version = None
        self.auth = None

        self.oauth = None
        self.client_id = None
        self.client_secret = None
        self.token = None

        self.proxy_dict = None

    def is_valid(self):
        valid = False

        if self.api_version == '1.0':
            valid = True if self.auth else False
        elif self.api_version == '2.0':
            valid = True if self.oauth else False

        return valid

    @staticmethod
    def login(username, password):
        """ Connect to office 365 using specified username and password

        :param username: username to login with
        :param password: password for authentication
        """
        connection = Connection()

        connection.api_version = '1.0'
        connection.auth = (username, password)
        return connection

    @staticmethod
    def oauth2(client_id, client_secret, store_token=True, token_path=None):
        """ Connect to office 365 using specified Open Authentication protocol

        :param client_id: application_id generated by https://apps.dev.microsoft.com when you register your app
        :param client_secret: secret password key generated for your application
        :param store_token: whether or not to store the token in file system, so u don't have to keep opening
            the auth link and authenticating every time
        :param token_path: full path to where the token should be saved to
        """
        connection = Connection()

        connection.api_version = '2.0'
        connection.client_id = client_id
        connection.client_secret = client_secret

        if not store_token:
            delete_token(token_path)

        token = load_token(token_path)

        if not token:
            connection.oauth = OAuth2Session(client_id=client_id,
                                             redirect_uri='https://outlook.office365.com/owa/',
                                             scope=['https://graph.microsoft.com/Mail.ReadWrite',
                                                    'https://graph.microsoft.com/Mail.Send',
                                                    'offline_access'], )
            oauth = connection.oauth
            auth_url, state = oauth.authorization_url(
                url=Connection._oauth2_authorize_url,
                access_type='offline')
            print('Please open {} and authorize the application'.format(auth_url))
            auth_resp = input('Enter the full result url: ')
            os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = 'Y'
            token = oauth.fetch_token(token_url=Connection._oauth2_token_url,
                                      authorization_response=auth_resp, client_secret=client_secret)
            save_token(token, token_path)
        else:
            connection.oauth = OAuth2Session(client_id=client_id,
                                             token=token)

        return connection

    @staticmethod
    def proxy(url, port, username, password):
        """ Connect to Office 365 though the specified proxy

        :param url: url of the proxy server
        :param port: port to connect to proxy server
        :param username: username for authentication in the proxy server
        :param password: password for the specified username
        """
        connection = Connection()

        connection.proxy_dict = {
            "http": "http://{}:{}@{}:{}".format(username, password, url, port),
            "https": "https://{}:{}@{}:{}".format(username, password, url,
                                                  port),
        }
        return connection

    @staticmethod
    def get_response(request_url, **kwargs):
        """ Fetches the response for specified url and arguments, adding the auth and proxy information to the url

        :param request_url: url to request
        :param kwargs: any keyword arguments to pass to the requests api
        :return: response object
        """
        connection = Connection()

        if not connection.is_valid():
            raise RuntimeError('Connection is not configured, please use "O365.Connection" '
                               'to set username and password or OAuth2 authentication')

        con_params = {}
        if connection.proxy_dict:
            con_params['proxies'] = connection.proxy_dict
        con_params.update(kwargs)

        log.info('Requesting URL: {}'.format(request_url))

        if connection.api_version == '1.0':
            con_params['auth'] = connection.auth
            response = requests.get(request_url, **con_params)
        else:
            try:
                response = connection.oauth.get(request_url, **con_params)
            except TokenExpiredError:
                log.info('Token is expired, fetching a new token')
                token = connection.oauth.refresh_token(Connection._oauth2_token_url, client_id=connection.client_id,
                                                       client_secret=connection.client_secret)
                log.info('New token fetched')
                save_token(token)

                response = connection.oauth.get(request_url, **con_params)

        log.info('Received response from URL {}'.format(response.url))

        response_json = response.json()
        if 'value' not in response_json:
            raise RuntimeError('Something went wrong, received an unexpected result \n{}'.format(response_json))

        response_values = [MicroDict(x) for x in response_json['value']]
        return response_values



from O365.contact import Contact
import logging
import json
import requests

log = logging.getLogger(__name__)

class Group( object ):
	'''
	A wrapper class that handles all the contacts associated with a single Office365 account.
	
	Methods:
		constructor -- takes your email and password for authentication.
		getContacts -- begins the actual process of downloading contacts.
	
	Variables:
		con_url -- the url that is requested for the retrival of the contacts.
		con_folder_url -- the url that is used for requesting contacts from a specific folder.
		folder_url -- the url that is used for finding folder Id's from folder names.
	'''
	con_url = 'https://outlook.office365.com/api/v1.0/me/contacts'
	con_folder_url = 'https://outlook.office365.com/api/v1.0/me/contactfolders/{0}/contacts'
	folder_url = 'https://outlook.office365.com/api/v1.0/me/contactfolders?$filter=DisplayName eq \'{0}\''

	def __init__(self, auth, folderName=None,verify=True):
		'''
		Creates a group class for managing all contacts associated with email+password.

		Optional: folderName -- send the name of a contacts folder and the search will limit
		it'self to only those which are in that folder.
		'''
		log.debug('setting up for the schedule of the email %s',auth[0])
		self.auth = auth
		self.contacts = []
		self.folderName = folderName

		self.verify = verify


	def getContacts(self):
		'''Begin the process of downloading contact metadata.'''
		if self.folderName is None:
			log.debug('fetching contacts.')
			response = requests.get(self.con_url,auth=self.auth,verify=self.verify)
			log.info('Response from O365: %s', str(response))

		else:
			log.debug('fetching contact folder.')
			response = requests.get(self.folder_url.format(self.folderName),auth=self.auth,verify=self.verify)
			fid = response.json()['value'][0]['Id']
			log.debug('got a response of {0} and an Id of {1}'.format(response.status_code,fid))

			log.debug('fetching contacts for {0}.'.format(self.folderName))
			response = requests.get(self.con_folder_url.format(fid),auth=self.auth,verify=self.verify)
			log.info('Response from O365: {0}'.format(str(response)))

		for contact in response.json()['value']:
			duplicate = False
			log.debug('Got a contact Named: {0}'.format(contact['DisplayName']))
			for existing in self.contacts:
				if existing.json['Id'] == contact['Id']:
					log.info('duplicate contact')
					duplicate = True
					break

			if not duplicate:
				self.contacts.append(Contact(contact,self.auth))
			
			log.debug('Appended Contact.')
				
			
		log.debug('all calendars retrieved and put in to the list.')
		return True

#To the King!



from O365.cal import Calendar
import logging
import json
import requests

log = logging.getLogger(__name__)

class Schedule( object ):
	'''
	A wrapper class that handles all the Calendars associated with a sngle Office365 account.
	
	Methods:
		constructor -- takes your email and password for authentication.
		getCalendars -- begins the actual process of downloading calendars.
	
	Variables:
		cal_url -- the url that is requested for the retrival of the calendar GUIDs.
	'''
	cal_url = 'https://outlook.office365.com/api/v1.0/me/calendars'

	def __init__(self, auth, verify=True):
		'''Creates a Schedule class for managing all calendars associated with email+password.'''
		log.debug('setting up for the schedule of the email %s',auth[0])
		self.auth = auth
		self.calendars = []

		self.verify = verify


	def getCalendars(self):
		'''Begin the process of downloading calendar metadata.'''
		log.debug('fetching calendars.')
		response = requests.get(self.cal_url,auth=self.auth,verify=self.verify)
		log.info('Response from O365: %s', str(response))
		
		for calendar in response.json()['value']:
			try:
				duplicate = False
				log.debug('Got a calendar with Name: {0} and Id: {1}'.format(calendar['Name'],calendar['Id']))
				for i,c in enumerate(self.calendars):
					if c.json['Id'] == calendar['Id']:
						c.json = calendar
						c.name = calendar['Name']
						c.calendarId = calendar['Id']
						duplicate = True
						log.debug('Calendar: {0} is a duplicate',calendar['Name'])
						break

				if not duplicate:
					self.calendars.append(Calendar(calendar,self.auth))
					log.debug('appended calendar: %s',calendar['Name'])

				log.debug('Finished with calendar {0} moving on.'.format(calendar['Name']))

			except Exception as e:
				log.info('failed to append calendar: {0}'.format(str(e)))
		
		log.debug('all calendars retrieved and put in to the list.')
		return True

#To the King!
